#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>


from ltp.patchs.pytorch_lightning.patch_1_1_to_1_2 import patch as pl_1_2_patch_1_1

from ltp.patchs.model.patch_4_1_3 import patch as model_patch_4_1_3
